import 'dart:async';

import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/scheduler.dart';
import 'package:flutter/widgets.dart';

import '../forme.dart';
import 'forme_field_scope.dart';

/// form key is a global key , also used to manage form
class FormeKey extends LabeledGlobalKey<State> implements FormeController {
  FormeKey({String? debugLabel}) : super(debugLabel);

  /// whether formKey is initialized
  bool get initialized {
    final State? state = currentState;
    if (state == null || state is! _FormeState) {
      return false;
    }
    return true;
  }

  /// get form controller , throw an error if there's no form controller
  FormeController get _currentController {
    final State? currentState = super.currentState;
    if (currentState == null) {
      throw Exception('current state is null,did you put this key on Forme?');
    }
    if (currentState is! _FormeState) {
      throw Exception(
          'current state is not a state of Forme , did you put this key on Forme?');
    }
    return currentState.controller;
  }

  @override
  Map<String, Object?> get value => _currentController.value;

  @override
  FormeValidation get validation => _currentController.validation;

  @override
  T field<T extends FormeFieldController<Object?>>(String name) =>
      _currentController.field<T>(name);

  @override
  bool hasField(String name) => _currentController.hasField(name);

  @override
  Future<FormeValidateSnapshot> validate({
    bool quietly = false,
    Set<String> names = const {},
    bool clearError = false,
    bool validateByOrder = false,
  }) =>
      _currentController.validate(
        quietly: quietly,
        names: names,
        validateByOrder: validateByOrder,
        clearError: clearError,
      );

  @override
  void reset() => _currentController.reset();

  @override
  void save() => _currentController.save();

  @override
  set value(Map<String, Object?> data) => _currentController.value = data;

  static FormeController? of(BuildContext context) {
    return _FormeScope.of(context)?.controller;
  }

  @override
  bool get quietlyValidate => _currentController.quietlyValidate;

  @override
  bool get isValueChanged => _currentController.isValueChanged;

  @override
  List<FormeFieldController> get controllers => _currentController.controllers;

  @override
  ValueListenable<FormeFieldController<T>?> fieldListenable<T>(String name) =>
      _currentController.fieldListenable<T>(name);

  @override
  ValueListenable<FormeValidation> get validationListenable =>
      _currentController.validationListenable;

  @override
  ValueListenable<Map<String, FormeFieldController?>> get fieldsListenable =>
      _currentController.fieldsListenable;

  @override
  bool addVisitor(FormeVisitor visitor) =>
      _currentController.addVisitor(visitor);

  @override
  bool removeVisitor(FormeVisitor visitor) =>
      _currentController.removeVisitor(visitor);
}

/// build your form !
class Forme extends StatefulWidget {
  /// listen form value changed
  ///
  /// this listener will be always triggered when field value changed
  final FormeValueChanged<Object?>? onValueChanged;

  /// listen form focus changed
  final FormeFocusChanged<Object?>? onFocusChanged;

  /// form content
  final Widget child;

  /// map initial value
  ///
  /// **this property can be overwritten by field's initialValue**
  final Map<String, Object?> initialValue;

  /// used to listen field's validation changed
  final FormeFieldValidationChanged<Object?>? onFieldValidationChanged;

  final WillPopCallback? onWillPop;

  /// if this flag is true , will not display default error when perform a validate
  final bool quietlyValidate;

  /// autovalidateMode
  ///
  /// if mode is [AutovalidateMode.onUserInteraction] , will validate and rebuild all fields which has a validator
  /// after field was interacted by user
  ///
  /// if mode is [AutovalidateMode.always] , will revalidated all value fields after called reset
  final AutovalidateMode autovalidateMode;

  /// validated by field order , only one field will be validated at a time.
  ///
  /// will not continue if any field validation is not passed
  final bool autovalidateByOrder;

  /// listen fields registered in this frame
  final void Function(List<FormeFieldController> controllers)?
      onFieldsRegistered;
  final void Function(List<String> names)? onFieldsUnregistered;

  /// listen [FormeValidation] changed
  final void Function(FormeController field, FormeValidation validation)?
      onValidationChanged;

  final void Function(FormeFieldController<Object?> field, bool readOnly)?
      onReadonlyChanged;
  final void Function(FormeFieldController<Object?> field, bool enable)?
      onEnabledChanged;

  const Forme({
    FormeKey? key,
    this.onValueChanged,
    required this.child,
    this.initialValue = const <String, Object?>{},
    this.onFieldValidationChanged,
    this.onValidationChanged,
    this.onWillPop,
    this.quietlyValidate = false,
    this.onFocusChanged,
    AutovalidateMode? autovalidateMode,
    this.autovalidateByOrder = false,
    this.onFieldsRegistered,
    this.onFieldsUnregistered,
    this.onReadonlyChanged,
    this.onEnabledChanged,
  })  : autovalidateMode = autovalidateMode ?? AutovalidateMode.disabled,
        super(key: key);

  @override
  State<Forme> createState() => _FormeState();

  static FormeController? of(BuildContext context) =>
      _FormeScope.of(context)?.controller;
}

class _FormeState extends State<Forme> {
  final List<FormeFieldState> states = [];
  late final _FormeController controller;
  final Map<String, ValueNotifier<FormeFieldController?>> fieldNotifiers = {};
  final ValueNotifier<Map<String, FormeFieldController?>> fieldsNotifier =
      FormeMountedValueNotifier({});
  final ValueNotifier<FormeValidation> validationNotifier =
      FormeMountedValueNotifier(const FormeValidation({}));
  final List<FormeFieldState> newRegisteredStates = [];
  final List<String> newUnregisteredStates = [];
  final List<FormeVisitor> visitors = [];

  Map<String, Object?> get initialValue => widget.initialValue;

  AutovalidateMode get autovalidateMode => widget.autovalidateMode;

  int gen = 0;

  bool get quietlyValidate => widget.quietlyValidate;

  FormeValidation get validation => FormeValidation(states
      .where((element) => element.enabled)
      .toList()
      .asMap()
      .map((key, value) => MapEntry(value.name, value._status.validation)));

  @override
  void initState() {
    super.initState();
    controller = _FormeController(this);
  }

  @override
  void dispose() {
    controller._state = null;
    visitors.clear();
    fieldsNotifier.dispose();
    validationNotifier.dispose();
    fieldNotifiers.forEach((key, value) {
      value.dispose();
    });
    fieldNotifiers.clear();
    newRegisteredStates.clear();
    newUnregisteredStates.clear();
    super.dispose();
  }

  void reset() {
    for (final FormeFieldState element in states) {
      element.reset();
    }
    if (widget.autovalidateMode == AutovalidateMode.always) {
      setState(() {
        ++gen;
      });
    }
  }

  ValueNotifier<FormeFieldController<T>?> fieldListenable<T extends Object?>(
      String name) {
    return (fieldNotifiers[name] ??
            fieldNotifiers.putIfAbsent(
                name,
                () => FormeMountedValueNotifier<FormeFieldController<T>?>(
                    getFieldController<T>(name))))
        as ValueNotifier<FormeFieldController<T>?>;
  }

  FormeFieldController<T>? getFieldController<T extends Object?>(String name) {
    final List<FormeFieldController> controllers = states
        .where((element) => element.name == name)
        .map((e) => e.controller)
        .toList();
    return controllers.isEmpty
        ? null
        : controllers.first as FormeFieldController<T>;
  }

  Future<FormeValidation> _validateForm() async {
    if (widget.autovalidateByOrder) {
      final List<FormeFieldState> valueFieldStates = states
          .where((element) => element._hasAnyValidator && element.enabled)
          .toList()
        ..sort((a, b) => a.order!.compareTo(b.order!));
      if (valueFieldStates.isEmpty) {
        return const FormeValidation({});
      }
      return _validateByOrder(valueFieldStates, collector: {});
    } else {
      final Map<String, FormeFieldValidation> validations = {};
      await Future.wait(states
          .where((element) => element.enabled)
          .map((e) => e._validateByForm().then((value) {
                validations[e.name] = value;
                return value;
              })));
      return FormeValidation(validations);
    }
  }

  Future<FormeValidation> _validateByOrder(List<FormeFieldState> states,
      {int index = 0,
      required Map<String, FormeFieldValidation> collector}) async {
    final int length = states.length;
    if (index >= length) {
      return FormeValidation(collector);
    }
    final FormeFieldState state = states[index];
    final FormeFieldValidation validation = await state._validateByForm();
    collector[state.name] = validation;
    if (validation.isUnnecessary || validation.isValid) {
      return await _validateByOrder(states,
          index: index + 1, collector: collector);
    }
    return FormeValidation(collector);
  }

  void save() {
    for (final FormeFieldState element in states) {
      element.save();
    }
  }

  @override
  Widget build(BuildContext context) {
    if (needValidate) {
      _validateForm();
    }
    return WillPopScope(
      onWillPop: widget.onWillPop,
      child: _FormeScope(
        gen: gen,
        state: this,
        child: widget.child,
      ),
    );
  }

  Object? getInitialValue(String name, Object? value) {
    if (widget.initialValue.containsKey(name)) {
      return widget.initialValue[name];
    }
    return value;
  }

  bool get needValidate {
    if (autovalidateMode == AutovalidateMode.always) {
      return true;
    }
    if (autovalidateMode == AutovalidateMode.onUserInteraction) {
      return states.any((element) => element._hasInteractedByUser);
    }
    return false;
  }

  void fieldDidChange() {
    if (needValidate) {
      setState(() {
        ++gen;
      });
    }
  }

  void updateValidation() {
    final FormeValidation validation = this.validation;
    widget.onValidationChanged?.call(controller, validation);
    validationNotifier.value = validation;
  }

  void notifyFieldsRegistered(List<FormeFieldState> states) {
    for (final FormeVisitor visitor in visitors) {
      visitor.onFieldsRegistered(
          controller, states.map((e) => e.controller).toList());
    }
    for (final FormeFieldState state in states) {
      fieldNotifiers[state.name]?.value = state.controller;
    }
    fieldsNotifier.value = states
        .asMap()
        .map((key, value) => MapEntry(value.name, value.controller));
    widget.onFieldsRegistered?.call(states.map((e) => e.controller).toList());
    updateValidation();
  }

  void notifyFieldsUnregistered(List<String> names) {
    for (final FormeVisitor visitor in visitors) {
      visitor.onFieldsUnregistered(controller, List.of(names));
    }
    for (final String name in names) {
      fieldNotifiers[name]?.value = null;
    }
    fieldsNotifier.value =
        names.asMap().map((key, value) => MapEntry(value, null));
    widget.onFieldsUnregistered?.call(names);
    updateValidation();
  }

  void notifiyFieldsStatusChanged(
    FormeFieldState state,
    FormeFieldStatus oldStatus,
    FormeFieldStatus newStatus,
  ) {
    if (!states.contains(state)) {
      return;
    }
    for (final FormeVisitor visitor in visitors) {
      visitor.onFieldStatusChanged(
          controller, state.controller, oldStatus, newStatus);
    }

    if (oldStatus.validation != newStatus.validation) {
      updateValidation();
      widget.onFieldValidationChanged
          ?.call(state.controller, newStatus.validation);
    }
    if (oldStatus.hasFocus != newStatus.hasFocus) {
      widget.onFocusChanged?.call(state.controller, newStatus.hasFocus);
    }
    if (oldStatus.value != newStatus.value) {
      widget.onValueChanged?.call(state.controller, newStatus.value);
    }
    if (oldStatus.readOnly != newStatus.readOnly) {
      widget.onReadonlyChanged?.call(state.controller, newStatus.readOnly);
    }
    if (oldStatus.enabled != newStatus.enabled) {
      widget.onEnabledChanged?.call(state.controller, newStatus.enabled);
    }
  }

  void registerField(FormeFieldState state) {
    if (!states.contains(state)) {
      states.add(state);
      if (newRegisteredStates.isEmpty) {
        SchedulerBinding.instance!.endOfFrame.then((_) {
          if (needValidate) {
            _validateForm();
          }
          final List<FormeFieldState> states = List.of(newRegisteredStates);
          newRegisteredStates.clear();
          notifyFieldsRegistered(states);
        });
      }
      newRegisteredStates.add(state);
    }
  }

  void unregisterField(FormeFieldState state) {
    if (states.remove(state)) {
      if (newUnregisteredStates.isEmpty) {
        SchedulerBinding.instance!.endOfFrame.then((_) {
          if (mounted) {
            if (needValidate) {
              _validateForm();
            }
            final List<String> names = List.of(newUnregisteredStates);
            newUnregisteredStates.clear();
            notifyFieldsUnregistered(names);
          }
        });
      }
      newUnregisteredStates.add(state.name);
    }
  }

  bool get isValueChanged => states.any((element) => element.isValueChanged);

  int getOrder(FormeFieldState formeFieldState) {
    return states.indexOf(formeFieldState);
  }
}

class _FormeScope extends InheritedWidget {
  final int gen;
  final _FormeState state;

  const _FormeScope({
    required this.gen,
    required this.state,
    required Widget child,
  }) : super(child: child);

  @override
  bool updateShouldNotify(covariant _FormeScope oldWidget) {
    return gen != oldWidget.gen;
  }

  static _FormeState? of(BuildContext context) {
    return context.dependOnInheritedWidgetOfExactType<_FormeScope>()?.state;
  }
}

class FormeFieldState<T extends Object?> extends State<FormeField<T>> {
  final Duration _defaultAsyncValidatorDebounce =
      const Duration(milliseconds: 500);

  FocusNode? _focusNode;
  Timer? _asyncValidatorTimer;
  bool _ignoreValidate = false;
  bool _hasInteractedByUser = false;
  int _validateGen = 0;

  late FormeFieldStatus<T> _status;
  _FormeState? _formeState;

  late FormeFieldController<T> controller;

  int? get order => widget.order ?? _formeState?.getOrder(this);

  String get name => widget.name;
  T get value => _status.value;

  bool _inited = false;

  bool get _hasValidator => widget.validator != null;
  bool get _hasAsyncValidator => widget.asyncValidator != null;
  bool get _hasAnyValidator => _hasValidator || _hasAsyncValidator;

  bool get readOnly => _status.readOnly;
  bool get enabled => _status.enabled;

  bool? _readOnly;
  bool? _enabled;

  bool get _isReadOnly => (_readOnly ?? widget.readOnly) || !_isEnabled;
  bool get _isEnabled => _enabled ?? widget.enabled;
  FormeFieldValidation get _validation {
    if (!_isEnabled || !_hasAnyValidator) {
      return FormeFieldValidation.unnecessary;
    }
    if (_hasAnyValidator &&
        _status.validation == FormeFieldValidation.unnecessary) {
      return FormeFieldValidation.waiting;
    }
    return _status.validation;
  }

  /// set field readonly or not
  set readOnly(bool readOnly) {
    _readOnly = readOnly;
    if (_isReadOnly != _status.readOnly) {
      setState(() {
        _status = _status._copyWith(readOnly: _Optional(_isReadOnly));
      });
    }
  }

  /// set field enable or not
  ///
  /// if field is disabled
  ///
  /// 1. field can not be focused
  /// 2. field's validation is unnecessary
  /// 3. field is readOnly
  /// 4. trigger [onReadonlyChanged] if readonly state changed
  /// 5. trigger onValidationChanged if validation changed
  /// 6. trigger onEnabledChanged
  set enabled(bool enabled) {
    _enabled = enabled;
    setState(() {
      _validateGen++;
      if (!_isEnabled) {
        _hasInteractedByUser = false;
      }
      _status = _status._copyWith(
        enabled: _Optional(_isEnabled),
        readOnly: _Optional(_isReadOnly),
        validation: _Optional(_validation),
      );
    });
  }

  FormeFieldValidation get _initialValidation => _hasAnyValidator && enabled
      ? FormeFieldValidation.waiting
      : FormeFieldValidation.unnecessary;

  String? get errorText => !enabled ||
          (_formeState?.quietlyValidate ?? false) ||
          widget.quietlyValidate
      ? null
      : _status.validation.error;

  /// get initialValue
  T get initialValue =>
      _formeState?.getInitialValue(name, widget.initialValue) as T ??
      widget.initialValue;

  /// get current widget's focus node
  ///
  /// if there's no focus node,will create a new one
  FocusNode get focusNode {
    if (_focusNode == null) {
      focusNode = _DisposeRequiredFocusNode();
      focusNode.canRequestFocus = enabled;
    }
    return _focusNode!;
  }

  /// **this method should be called by subclass only!**
  ///
  /// if current focusNode is not null ,dispose current node & set new focusNode
  ///
  /// you need to dispose node you set before before you set another node
  @protected
  set focusNode(FocusNode focusNode) {
    if (_focusNode == focusNode) {
      return;
    }
    if (_focusNode is _DisposeRequiredFocusNode) {
      _focusNode!.dispose();
    }
    _focusNode = focusNode;
    if (!enabled) {
      _focusNode!.canRequestFocus = false;
    }
    _focusNode!.addListener(_onFocusChangedListener);
  }

  void _onFocusChangedListener() {
    final FormeFieldStatus<T> old = _status;
    _status = _status._copyWith(
      hasFocus: _Optional(_focusNode?.hasFocus ?? false),
    );
    _onModelChanged(old, _status);
  }

  bool get isValueChanged => initialValue != value;

  void _register() {
    if (widget.registrable) {
      _formeState?.registerField(this);
    }
  }

  /// if you want to init some resources relies on [initialValue] or [readOnly] or [enabled]
  ///
  /// use [initModel] instead
  @override
  void initState() {
    super.initState();

    if (widget.onInitialed != null) {
      WidgetsBinding.instance!.addPostFrameCallback((timeStamp) {
        widget.onInitialed?.call(controller);
      });
    }
  }

  @override
  void didUpdateWidget(covariant FormeField<T> oldWidget) {
    super.didUpdateWidget(oldWidget);

    if (widget.registrable) {
      _formeState?.registerField(this);
    } else {
      _formeState?.unregisterField(this);
    }

    final FormeFieldStatus<T> old = _status;
    _status = _status._copyWith(
      readOnly: _Optional(_isReadOnly),
      enabled: _Optional(_isEnabled),
      validation: _Optional(_validation),
    );

    if (widget.valueUpdater != null) {
      T newValue = widget.valueUpdater!(oldWidget, widget, value);
      _status = _status._copyWith(
        value: _Optional(newValue),
      );
    }

    if (old.validation != _status.validation) {
      _validateGen++;
    }

    _onModelChanged(old, _status, true);
  }

  @override
  void setState(VoidCallback fn) {
    final FormeFieldStatus<T> old = _status;
    super.setState(fn);
    _onModelChanged(old, _status);
  }

  @override
  @mustCallSuper
  void didChangeDependencies() {
    super.didChangeDependencies();
    _formeState = _FormeScope.of(context);

    if (!_inited) {
      _inited = true;

      initModel();
      controller = createFormeFieldController();
    } else {
      final FormeFieldStatus<T> old = _status;
      _status = _status._copyWith(
        readOnly: _Optional(_isReadOnly),
      );
      _onModelChanged(old, _status, true);
    }
  }

  /// this method will be called only once in state's lifecircle,immediately  called after [initState]
  ///
  /// recommend to init your resources in this method rather than [initState]
  ///
  /// Implementations of this method should start with a call to the inherited
  /// method
  @protected
  @mustCallSuper
  void initModel() {
    _status = FormeFieldStatus<T>._(
      enabled: widget.enabled,
      readOnly: widget.readOnly || !widget.enabled,
      validation: _hasAnyValidator && widget.enabled
          ? FormeFieldValidation.waiting
          : FormeFieldValidation.unnecessary,
      value: initialValue,
      hasFocus: _focusNode?.hasFocus ?? false,
      disposed: false,
    );
  }

  /// create [FormeFieldController] , this method will only called once in field's lifecycle
  ///
  /// if you want to override this method, use [FormeFieldControllerDelegate] to wrap parent's controller
  @protected
  FormeFieldController<T> createFormeFieldController() =>
      _FormeFieldController(_status, this);

  @override
  void deactivate() {
    _formeState?.unregisterField(this);
    super.deactivate();
  }

  @override
  void dispose() {
    controller.statusNotifier.value =
        _status._copyWith(disposed: _Optional(true));
    _asyncValidatorTimer?.cancel();
    _focusNode?.removeListener(_onFocusChangedListener);
    if (_focusNode is _DisposeRequiredFocusNode) {
      _focusNode?.dispose();
    }
    super.dispose();
  }

  @mustCallSuper
  void didChange(T newValue) {
    if (_status.value != newValue) {
      setState(() {
        _hasInteractedByUser = true;
        _status = _status._copyWith(value: _Optional(newValue));
      });
      _fieldChange();
    }
  }

  @mustCallSuper
  void reset() {
    setState(() {
      _validateGen++;
      _hasInteractedByUser = false;
      _ignoreValidate = false;
      _status = _status._copyWith(
          validation: _Optional(_initialValidation),
          value: _Optional(initialValue));
    });
    _fieldChange();
  }

  set value(T value) => didChange(value);

  @override
  Widget build(BuildContext context) {
    final bool needValidate = _hasAnyValidator &&
        enabled &&
        ((widget.autovalidateMode == AutovalidateMode.always) ||
            (widget.autovalidateMode == AutovalidateMode.onUserInteraction &&
                _hasInteractedByUser));
    if (needValidate) {
      _validateByField();
    }

    Widget child = widget.builder(this);

    if (widget.decorator != null) {
      child = widget.decorator!.build(
        controller,
        child,
      );
    }

    _register();
    return FormeFieldScope(controller, child);
  }

  void _onModelChanged(
      FormeFieldStatus<T> oldStatus, FormeFieldStatus<T> newStatus,
      [bool onlyAfterFrameCompleted = false]) {
    void _perform(VoidCallback fn) {
      void task() {
        controller.statusNotifier.value = newStatus;
        _formeState?.notifiyFieldsStatusChanged(this, oldStatus, newStatus);
        fn();
      }

      if (onlyAfterFrameCompleted) {
        WidgetsBinding.instance!.addPostFrameCallback((timeStamp) {
          task();
        });
      } else {
        task();
      }
    }

    if (oldStatus.enabled != newStatus.enabled) {
      _focusNode?.canRequestFocus = newStatus.enabled;
      _perform(() {
        widget.onEnabledChanged?.call(controller, newStatus.enabled);
        onEnabledChanged(newStatus.enabled);
      });
    }
    if (oldStatus.readOnly != newStatus.readOnly) {
      _perform(() {
        widget.onReadonlyChanged?.call(controller, newStatus.readOnly);
        onReadonlyChanged(newStatus.readOnly);
      });
    }
    if (oldStatus.validation != newStatus.validation) {
      _perform(() {
        widget.onValidationChanged?.call(controller, newStatus.validation);
        onValidationChanged(newStatus.validation);
      });
    }
    if (oldStatus.value != newStatus.value) {
      _ignoreValidate = false;
      _perform(() {
        widget.onValueChanged?.call(controller, newStatus.value);
        onValueChanged(newStatus.value);
      });
    }

    if (oldStatus.hasFocus != newStatus.hasFocus) {
      controller.statusNotifier.value = newStatus;
      widget.onFocusChanged?.call(controller, newStatus.hasFocus);
      onFocusChanged(newStatus.hasFocus);
    }
  }

  /// override this method if you want to listen value changed
  @protected
  void onValueChanged(T value) {}

  /// override this method if you want to listen validation changed
  @protected
  void onValidationChanged(FormeFieldValidation validation) {}

  /// override this method if you want to listen focus changed
  @protected
  void onFocusChanged(bool hasFocus) {}

  /// override this method if you want to listen readonly-state change
  @protected
  void onReadonlyChanged(bool readOnly) {}

  /// override this method if you want to listen enable-state change
  @protected
  void onEnabledChanged(bool enable) {}

  void requestFocusOnUserInteraction() {
    if (_hasInteractedByUser && widget.requestFocusOnUserInteraction) {
      _focusNode?.requestFocus();
    }
  }

  void save() {
    widget.onSaved?.call(controller, value);
  }

  void _clearError() {
    if (_status.validation != _initialValidation) {
      setState(() {
        _ignoreValidate = false;
        _validateGen++;
        _status = _status._copyWith(validation: _Optional(_initialValidation));
      });
    }
  }

  /// this method should be only called in [_FormeState.build]
  Future<FormeFieldValidation> _validateByForm() async {
    void notifyValidation(FormeFieldValidation validation) {
      if (mounted && validation != _status.validation) {
        setState(() {
          _status = _status._copyWith(validation: _Optional(validation));
        });
      }
    }

    if (_ignoreValidate || !_hasAnyValidator) {
      if (!_hasAnyValidator) {
        notifyValidation(FormeFieldValidation.unnecessary);
      }
      return _status.validation;
    }

    final int gen = ++_validateGen;
    if (_hasValidator) {
      final String? errorText = widget.validator!(controller, value);
      if (errorText != null || !_hasAsyncValidator) {
        notifyValidation(_createFormeFieldValidation(errorText));
        return _status.validation;
      }
    }

    if (_hasAsyncValidator) {
      notifyValidation(FormeFieldValidation.validating);
      await _performAsyncValidate(gen);
    }
    return _status.validation;
  }

  /// this method should  be only called in [FormeFieldState.build]
  void _validateByField() {
    void notifyValidation(FormeFieldValidation validation) {
      if (mounted && _status.validation != validation) {
        final FormeFieldStatus<T> oldStatus = _status;
        _status = _status._copyWith(validation: _Optional(validation));
        _onModelChanged(oldStatus, _status, true);
      }
    }

    if (_ignoreValidate || !_hasAnyValidator) {
      if (!_hasAnyValidator) {
        notifyValidation(FormeFieldValidation.unnecessary);
      }
      return;
    }

    final int gen = ++_validateGen;

    if (_hasValidator) {
      final String? errorText = widget.validator!(controller, value);
      if (errorText != null || !_hasAsyncValidator) {
        notifyValidation(_createFormeFieldValidation(errorText));
        return;
      }
    }
    if (_hasAsyncValidator) {
      notifyValidation(FormeFieldValidation.validating);
      _asyncValidate(gen);
    }
  }

  void _asyncValidate(
    int gen, {
    VoidCallback? onCompleted,
  }) {
    _asyncValidatorTimer?.cancel();
    _asyncValidatorTimer =
        Timer(widget.asyncValidatorDebounce ?? _defaultAsyncValidatorDebounce,
            () async {
      final bool isValid = await _performAsyncValidate(gen);
      if (isValid) {
        onCompleted?.call();
      }
    });
  }

  Future<bool> _performAsyncValidate(int gen) async {
    bool isValid() {
      return mounted && gen == _validateGen;
    }

    FormeFieldValidation validation;

    try {
      final String? errorText =
          await widget.asyncValidator!(controller, value, isValid);
      validation = _createFormeFieldValidation(errorText);
    } catch (e, stackTrace) {
      validation = FormeFieldValidation.fail(e, stackTrace);
    }

    if (isValid()) {
      setState(() {
        _ignoreValidate = true;
        _status = _status._copyWith(validation: _Optional(validation));
      });
      return true;
    }
    return false;
  }

  void _fieldChange() {
    _formeState?.fieldDidChange();
  }

  FormeFieldValidation _createFormeFieldValidation(String? errorText) {
    if (errorText == null) {
      return FormeFieldValidation.valid;
    }
    return FormeFieldValidation.invalid(errorText);
  }

  /// this method is used to manually validate
  Future<FormeFieldValidateSnapshot<T>> _validateManually(
      {bool quietly = false}) async {
    final T value = this.value;
    if (!_hasAnyValidator || !enabled) {
      return FormeFieldValidateSnapshot(
          value, FormeFieldValidation.unnecessary, controller, false, false);
    }
    final int gen = quietly ? _validateGen : ++_validateGen;

    bool isValid() {
      return mounted && gen == _validateGen;
    }

    bool needNotify() {
      return !quietly && isValid();
    }

    void notify(FormeFieldValidation validation) {
      if (needNotify() && _status.validation != validation) {
        setState(() {
          _status = _status._copyWith(validation: _Optional(validation));
        });
      }
    }

    if (_hasValidator) {
      final String? errorText = widget.validator!(controller, value);
      if (errorText != null || !_hasAsyncValidator) {
        final FormeFieldValidation validation =
            _createFormeFieldValidation(errorText);
        notify(validation);
        return FormeFieldValidateSnapshot(
            value, validation, controller, false, false);
      }
    }

    notify(FormeFieldValidation.validating);

    FormeFieldValidation validation;
    try {
      final String? errorText =
          await widget.asyncValidator!(controller, value, isValid);
      validation = _createFormeFieldValidation(errorText);
    } catch (e, stackTrace) {
      validation = FormeFieldValidation.fail(e, stackTrace);
    }

    notify(validation);
    return FormeFieldValidateSnapshot(value, validation, controller,
        value != this.value, value != initialValue);
  }

  void _markNeedsBuild() {
    setState(() {});
  }
}

class _FormeController extends FormeController {
  _FormeState? _state;
  @override
  final ValueListenable<Map<String, FormeFieldController?>> fieldsListenable;
  @override
  final ValueListenable<FormeValidation> validationListenable;
  _FormeController(this._state)
      : fieldsListenable = FormeValueListenableDelegate(_state!.fieldsNotifier),
        validationListenable =
            FormeValueListenableDelegate(_state.validationNotifier);

  _FormeState get state =>
      _state ??
      (throw Exception('controller is disposed , should not used anymore'));

  @override
  Map<String, Object?> get value {
    final Map<String, Object?> map = <String, Object?>{};
    for (final FormeFieldState element in state.states) {
      if (!element.enabled) {
        continue;
      }
      final String name = element.name;
      final Object? value = element.value;
      map[name] = value;
    }
    return map;
  }

  @override
  bool get quietlyValidate => state.quietlyValidate;

  @override
  FormeValidation get validation => state.validation;

  @override
  T field<T extends FormeFieldController<Object?>>(String name) {
    final T? field = findFormeFieldController(name);
    if (field == null) {
      throw Exception('no field can be found by name :$name');
    }
    return field;
  }

  @override
  bool hasField(String name) =>
      state.states.any((element) => element.name == name);

  @override
  Future<FormeValidateSnapshot> validate({
    bool quietly = false,
    Set<String> names = const {},
    bool clearError = false,
    bool validateByOrder = false,
  }) async {
    final List<FormeFieldState> states = (state.states
            .where((element) =>
                element._hasAnyValidator &&
                element.enabled &&
                (names.isEmpty || names.contains(element.name)))
            .toList()
          ..sort((a, b) => a.order!.compareTo(b.order!)))
        .toList();
    if (states.isEmpty) {
      return FormeValidateSnapshot([]);
    }
    if (clearError) {
      for (final FormeFieldState element in states) {
        element._clearError();
      }
    }
    if (validateByOrder) {
      return _validateByOrder(states, quietly);
    }
    final List<FormeFieldValidateSnapshot<Object?>> value = await Future.wait(
        states.map((state) => state._validateManually(quietly: quietly)),
        eagerError: true);
    return FormeValidateSnapshot(value);
  }

  Future<FormeValidateSnapshot> _validateByOrder(
      List<FormeFieldState> states, bool quietly,
      {int index = 0, List<FormeFieldValidateSnapshot> list = const []}) async {
    final int length = controllers.length;
    final List<FormeFieldValidateSnapshot> copyList = List.of(list);
    final FormeFieldValidateSnapshot snapshot =
        await states[index]._validateManually(quietly: quietly);
    copyList.add(snapshot);
    if (!snapshot.isValid || index == length - 1) {
      return FormeValidateSnapshot(copyList);
    }
    return _validateByOrder(states, quietly, index: index + 1, list: copyList);
  }

  @override
  void reset() => state.reset();

  @override
  void save() => state.save();

  @override
  set value(Map<String, Object?> data) => data.forEach((key, Object? value) {
        field(key).value = value;
      });

  T? findFormeFieldController<T extends FormeFieldController<Object?>>(
      String name) {
    for (final FormeFieldState state in state.states) {
      if (state.name == name) {
        return state.controller as T;
      }
    }
    return null;
  }

  @override
  bool get isValueChanged => state.isValueChanged;

  @override
  List<FormeFieldController> get controllers =>
      state.states.map((e) => e.controller).toList();

  @override
  ValueListenable<FormeFieldController<T>?> fieldListenable<T>(String name) =>
      FormeValueListenableDelegate<FormeFieldController<T>?>(
          state.fieldListenable<T>(name));

  @override
  bool addVisitor(FormeVisitor visitor) {
    if (!state.visitors.contains(visitor)) {
      state.visitors.add(visitor);
      return true;
    }
    return false;
  }

  @override
  bool removeVisitor(FormeVisitor visitor) {
    return state.visitors.remove(visitor);
  }
}

class _FormeFieldController<T extends Object?> extends FormeFieldController<T> {
  FormeFieldState<T>? _fieldState;

  @override
  final String name;
  @override
  final _FormeFieldStatusNotifier<T> statusNotifier;

  _FormeFieldController(FormeFieldStatus<T> status, FormeFieldState<T> state)
      : _fieldState = state,
        statusNotifier = _FormeFieldStatusNotifier(status),
        name = state.name {
    statusNotifier.addListener(() {
      final FormeFieldStatus<T> status = statusNotifier.value;
      if (status.disposed) {
        _fieldState = null;
        statusNotifier.dispose();
      }
    });
  }

  FormeFieldState<T> get _state =>
      _fieldState ??
      (throw Exception('controller is disposed , should not used anymore'));

  @override
  FocusNode? get focusNode => _state._focusNode;

  @override
  set readOnly(bool readOnly) => _state.readOnly = readOnly;

  @override
  void reset() => _state.reset();

  @override
  Future<FormeFieldValidateSnapshot<T>> validate({bool quietly = false}) =>
      _state._validateManually(quietly: quietly);

  @override
  set value(T value) => _state.value = value;

  @override
  bool get isValueChanged => _state.isValueChanged;

  @override
  set enabled(bool enabled) => _state.enabled = enabled;

  @override
  void markNeedsBuild() => _state._markNeedsBuild();

  @override
  ValueListenable<bool> get enabledListenable =>
      FormeValueListenableDelegate(statusNotifier.enabledListenable);

  @override
  ValueListenable<bool> get focusListenable =>
      FormeValueListenableDelegate(statusNotifier.focusListenable);

  @override
  T? get oldValue => statusNotifier.oldValue;

  @override
  ValueListenable<bool> get readOnlyListenable =>
      FormeValueListenableDelegate(statusNotifier.readOnlyListenable);

  @override
  ValueListenable<FormeFieldValidation> get validationListenable =>
      FormeValueListenableDelegate(statusNotifier.validationListenable);

  @override
  ValueListenable<T> get valueListenable =>
      FormeValueListenableDelegate(statusNotifier.valueListenable);

  @override
  BuildContext get context => _state.context;

  @override
  void clearError() => _state._clearError();

  @override
  int? get order => _state.order;
}

/// a focusnode created by FormeField itself rather than set by subclass ,
/// so it's our responsibility to dispose it
class _DisposeRequiredFocusNode extends FocusNode {}

class FormeFieldStatus<T extends Object?> {
  final bool enabled;
  final bool readOnly;
  final FormeFieldValidation validation;
  final T value;
  final bool hasFocus;

  /// field is disposed
  final bool disposed;

  FormeFieldStatus._({
    required this.enabled,
    required this.readOnly,
    required this.validation,
    required this.value,
    required this.hasFocus,
    required this.disposed,
  });

  FormeFieldStatus<T> _copyWith({
    _Optional<bool>? enabled,
    _Optional<bool>? readOnly,
    _Optional<FormeFieldValidation>? validation,
    _Optional<T>? value,
    _Optional<bool>? hasFocus,
    _Optional<bool>? disposed,
  }) {
    return FormeFieldStatus<T>._(
      enabled: enabled == null ? this.enabled : enabled.value,
      readOnly: readOnly == null ? this.readOnly : readOnly.value,
      validation: validation == null ? this.validation : validation.value,
      value: value == null ? this.value : value.value,
      hasFocus: hasFocus == null ? this.hasFocus : hasFocus.value,
      disposed: disposed == null ? this.disposed : disposed.value,
    );
  }
}

class _Optional<T> {
  final T value;
  _Optional(this.value);
}

class _FormeFieldStatusNotifier<T extends Object?>
    extends FormeMountedValueNotifier<FormeFieldStatus<T>> {
  T? oldValue;
  final ValueNotifier<bool> focusListenable;
  final ValueNotifier<bool> readOnlyListenable;
  final ValueNotifier<bool> enabledListenable;
  final ValueNotifier<T> valueListenable;
  final ValueNotifier<FormeFieldValidation> validationListenable;
  _FormeFieldStatusNotifier(
    FormeFieldStatus<T> value,
  )   : focusListenable = FormeMountedValueNotifier(value.hasFocus),
        readOnlyListenable = FormeMountedValueNotifier(value.readOnly),
        enabledListenable = FormeMountedValueNotifier(value.enabled),
        valueListenable = FormeMountedValueNotifier(value.value),
        validationListenable = FormeMountedValueNotifier(value.validation),
        super(value);

  @override
  set value(FormeFieldStatus<T> newValue) {
    super.value = newValue;
    if (!newValue.disposed) {
      readOnlyListenable.value = newValue.readOnly;
      focusListenable.value = newValue.hasFocus;
      enabledListenable.value = newValue.enabled;
      if (valueListenable.value != newValue.value) {
        oldValue = valueListenable.value;
        valueListenable.value = newValue.value;
      }
      validationListenable.value = newValue.validation;
    }
  }

  @override
  void dispose() {
    focusListenable.dispose();
    readOnlyListenable.dispose();
    enabledListenable.dispose();
    valueListenable.dispose();
    validationListenable.dispose();
    super.dispose();
  }
}
